(*
 * Abstract C expressions (the very small fragment of C to which we compile).
 *)

val cc : bool ref

type var =
  Var.var

type kind = EXP | STATEMENT | NEXP | NSTATEMENT

(*
Note: NEXP and NSTATEMENT are *not* kinds of C expressions but invocations 
of the runtime system functions for compilation of Nomadic Pict programs to a 
native code, where system functions were supposed to be encoded by using Pict 
C libraries. The idea of compilation to native code of Nomadic Pict programs 
was not realised though. 
*)

type info = {
  alloc: int;       (* Number of words allocated by the C code *)
  const: bool;      (* True if the C code is a constant *)
  reader: bool;     (* True if the C code reads from updatable storage *)
  writer: bool;     (* True if the C code writes to updatable storage *)
  kind: kind;       (* Tells us what kind of expression this is *)
  code: string list (* The list of pieces of C code *)
}

type ty =
  INTEGER
| POINTER

type decl =
  BLOCK of var * (var * ty) list * var * var * int * code
| CONST of var * exp list
| BYTES of var * string
| TOPLEVEL of string
| FORWARD of var
| EXTERN of var

and code =
  SEQ of code * code
| ASSIGN of exp * exp
| IF of exp * code * code
| CPROC of info * exp list
| PROFILING of string
| NULL

and exp =
  INT of int
| VAR of var
| ADDR of var
| OFFSET of bool * exp * int
| DEREF of var * int
| TAG of var * int
| TAGS of var
| INDEX of var * exp
| CCODE of info * exp list

val printDecls : out_channel -> decl list -> unit
val formatDecls : decl list -> unit

val caml_formatInfo : info -> unit

(*
 * Various C code fragments used in the code generator.
 *)

val complexInput : info
val complexOutput : info
val sizeOf : info
val free : info
val startq : info
val endq : info

