(*
 * Printing a fully-typed language for debugging (O'Caml format).
 *)

open Simplify
  
let caml_formatTy = function
    BOXED -> Format.print_string "BOXED"
  | UNBOXED -> Format.print_string "UNBOXED"
  | VAR(x) -> 
      Format.print_string "VAR("; Var.format x; 
      Format.print_char ')'

let rec caml_formatPat = function
    RECORDp(l) -> 
      let caml_formatFieldPat = function
	  ANONp(pat) -> 
	    Format.print_string "ANONp("; caml_formatPat pat; 
	    Format.print_char ')'
	| TYp(x) -> 
	    Format.print_string "TYp("; Var.format x;  
	    Format.print_char ')'
      in
      Format.print_string "RECORDp("; 
      Misc.formatList " [" "]" ";" caml_formatFieldPat l; 
      Format.print_char ')'; 
  | ASp(x,ty,pat) ->
      Format.print_string "ASp("; Var.format x;  
      Format.print_char ','; caml_formatTy ty; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ')'
  | VARp(x,ty) ->
      Format.print_string "VARp("; Var.format x;  
      Format.print_char ','; caml_formatTy ty; Format.print_char ')'
  | WILDp -> 
      Format.print_string "WILDp"

let rec caml_formatProc = function
    PRL(p,q) ->
      Format.print_string "PRL("; Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc q; 
      Format.print_char ')'
  | INPUT(v,pat,p) ->
      Format.print_string "INPUT("; caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | INPA(v,pat,p) ->
(*
  | INPA(pat,p) ->
  *)
      Format.print_string "INPA("; caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | OUTPUT(UNKNOWN,v,av) ->
      Format.print_string "OUTPUT(UNKNOWN,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal av; Format.print_char ')'
  | OUTPUT(RESPONSIVE,v,av) ->
      Format.print_string "OUTPUT(RESPONSIVE,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal av; Format.print_char ')'
  | OUTA(UNKNOWN,v,av) ->
      Format.print_string "OUTA(UNKNOWN,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal av; Format.print_char ')'
  | OUTA(RESPONSIVE,v,av) ->
      Format.print_string "OUTA(RESPONSIVE,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal av; Format.print_char ')'
  | LDOUTA(UNKNOWN,v,sv,av) ->
      Format.print_string "LDOUTA(UNKNOWN,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal sv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | LDOUTA(RESPONSIVE,v,sv,av) -> 
      Format.print_string "LDOUTA(RESPONSIVE,"; caml_formatVal v; 
      Format.print_char ','; caml_formatVal sv; Format.print_char ','; 
      caml_formatVal av; Format.print_char ')'
  | TRYHERE(v,av,p,q) -> 
      Format.print_string "TRYHERE("; caml_formatVal v; 
      Format.print_char ','; caml_formatVal av; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | REP_INPUT(v,pat,p) ->
      Format.print_string "REP_INPUT("; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | TIMED_INPUT(w,v,pat,p,q) ->
      Format.print_string "TIMED_INPUT("; 
      caml_formatVal w; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | REP_TIMED_INPUT(w,v,pat,p,q) ->
      Format.print_string "REP_TIMED_INPUT("; 
      caml_formatVal w; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | LI_SEND(v,cv,av) ->
      Format.print_string "LI_SEND(";
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | REMOTE_SEND(v,sv,cv,av) ->
      Format.print_string "REMOTE_SEND("; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal sv; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | LOCAL_SEND(v,cv,av) ->
      Format.print_string "LOCAL_SEND(";
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | IFLOCAL(v,cv,av,p,q) -> 
      Format.print_string "IFLOCAL("; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | SKIP -> 
      Format.print_string "SKIP"
  | LET(d,p) -> 
      Format.print_string "LET("; Format.print_break 1 2; caml_formatDec d; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ')'
  | IF(v,p,q) ->
      Format.print_string "IF("; Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc q;  Format.print_char ')'
  | TERMINATE -> 
      Format.print_string "TERMINATE"
  | TYPECASE(v, ts, p) -> 
      let caml_formatTCase l =
      	let caml_formatTCase (pat,t,p) = 
  	  Format.print_char '('; caml_formatPat pat; Format.print_char ',';
	  Types.caml_formatT t; Format.print_char ',';
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
      	      if needSep then begin
	  	Format.print_char ';';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatTCase h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatTCase h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "TYPECASE("; Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); caml_formatTCase ts; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ')'
  | CASE(v, cs) ->
      let caml_formatCase l =
      	let caml_formatCase (x,pat,p) = 
	  Format.print_char '('; Var.format x;  
	  Format.print_char ','; caml_formatPat pat; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
      	      if needSep then begin
	  	Format.print_char ';';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCase h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCase h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CASE("; Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); caml_formatCase cs;
      Format.print_char ')'
	
and caml_formatBoxProc p = 
  Format.open_hvbox 0; caml_formatProc p; Format.close_box()

and caml_formatDec = function
    SEQ(d1,d2) ->
      Format.print_string "SEQ("; Format.print_break 1 2; caml_formatDec d1; 
      Format.print_char ','; Format.print_space(); caml_formatDec d2; 
      Format.print_char ')'
  | DEF(l) -> 
      let caml_formatDef l =
       	let caml_formatDef (x,pat,p) =
  	  Format.print_char '('; Var.format x;  
	  Format.print_char ','; caml_formatPat pat; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatDef h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatDef h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "DEF("; caml_formatDef l; Format.print_char ')'
  | INLINE(n,pat,q) ->
      Format.print_string "INLINE("; Var.format n; Format.print_char ',';
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatProc q; Format.print_space(); Format.print_char ')'
  | VAL(pat,v) -> 
      Format.print_string "VAL("; caml_formatPat pat; Format.print_char ','; 
      caml_formatVal v; Format.print_char ')'
  | NEW(n,t) -> 
      Format.print_string "NEW("; Var.format n; Format.print_char ','; 
      caml_formatTy t; Format.print_char ')'
  | RUN(q) -> 
      Format.print_string "RUN("; Format.print_break 1 2; caml_formatBoxProc q; 
      Format.print_char ')'
  | NEWLOC(n) ->
      Format.print_string "NEWLOC("; Var.format n;  
      Format.print_char ')'
  | CREATE(l) -> 
      let caml_formatCreate l =
      	let caml_formatCreate (x,p) =
  	  Format.print_char '('; Var.format x; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CREATE("; caml_formatCreate l; Format.print_char ')'
  | CREATE_AGENT(l) -> 
      let caml_formatCreate l =
      	let caml_formatCreate (x,p) =
  	  Format.print_char '('; Var.format x; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CREATE_AGENT("; caml_formatCreate l; Format.print_char ')'
  | AT(a) -> 
      Format.print_string "AT("; Inter.caml_formatAtom a; Format.print_char ')'
  | MIGRATE(v) -> 
      Format.print_string "MIGRATE("; 
      (* caml_formatVal a; *) caml_formatVal v; 
      Format.print_char ')'

and caml_formatVal = function
    CCODEv(ci,l) ->
      Format.open_hovbox 0;  
      Format.print_string "CCODEv("; Ccode.caml_formatInfo ci; Format.print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatFieldVal l;
      Format.close_box(); Format.print_char ')'
  | RECORDv([]) ->
      Format.print_string "RECORDv([])"
  | RECORDv(l) ->
      Format.print_string "RECORDv("; 
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | APPv(s,v,al) ->
      begin
      	match s with
          RESPONSIVE -> 
	    Format.print_string "APPv(RESPONSIVE,"; 
	    caml_formatVal v; Format.print_char ','; 
	    Misc.formatList "[" "]" "; " caml_formatFieldVal al; 
	    Format.print_char ')'
      	| UNKNOWN -> 
	    Format.print_string "APPv(UNKNOWN,"; 
	    caml_formatVal v; Format.print_char ','; 
            Misc.formatList "[" "]" "; " caml_formatFieldVal al; 
	    Format.print_char ')'
      end
  | SELECTv(v,i) ->
      Format.print_string "SELECTv("; caml_formatVal v; Format.print_char ','; 
      Format.print_int i; Format.print_char ')'
  | STRINGv(s) ->
      Format.print_string "STRINGv(\""; Format.print_string s; 
      Format.print_string "\")"
  | SITEv(s,i) ->
      Format.print_string "SITEv(\""; Format.print_string s; 
      Format.print_int i; Format.print_string "\")"
  | CHARv(x) ->
      Format.print_string "CHARv(\'"; Format.print_char x; 
      Format.print_string "\')"
  | VARv(a) ->  
      Format.print_string "VARv("; Inter.caml_formatAtom a; Format.print_char ')'
  | BOOLv(b) ->
      Format.print_string "BOOLv("; Format.print_bool b; Format.print_char ')'
  | INTv(i) ->
      Format.print_string "INTv("; Format.print_int i; Format.print_char ')'
  | LETv(d,v) ->
      Format.print_string "LETv("; caml_formatDec d; Format.print_char ','; 
      caml_formatVal v; Format.print_char ')'
  | ABSv(pat,p) ->
      Format.print_string "ABSv("; caml_formatPat pat; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  | IFv(b,v1,v2) ->
      Format.print_string "IFv("; caml_formatVal b; Format.print_char ','; 
      caml_formatVal v1; Format.print_char ','; caml_formatVal v2; 
      Format.print_char ')'
  | DYNAMICv(v,t) ->
      Format.print_string "DYNAMICv("; caml_formatVal v; Format.print_char ','; 
      Types.caml_formatT t; Format.print_char ')'
  | TYPECASEv(v, ts, velse) ->
      let caml_formatTCasev (pat,t,v) = 
	Format.print_char '('; caml_formatPat pat; Format.print_char ',';
 	Types.caml_formatT t; Format.print_char ',';
	caml_formatVal v; Format.print_char ')' 
      in
      Format.print_string "TYPECASEv("; caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatTCasev ts;
      Format.print_char ','; caml_formatVal velse ; Format.print_char ')'
  | VARIANTv(n,v) ->
      Format.print_string "VARIANTv("; Var.format n; 
      Format.print_char ','; caml_formatVal v; Format.print_char ')'
  | CASEv(v, cs) ->
      let caml_formatCasev (n, pat,v) = 
	Format.print_char '('; Var.format n; Format.print_char ','; 
	caml_formatPat pat; Format.print_char ','; caml_formatVal v; 
	Format.print_char ')' 
      in
      Format.print_string "CASEv("; caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatCasev cs; 
      Format.print_char ')'
	
and caml_formatFieldVal = function
    ANONv(v) -> 
      Format.print_string "ANONv("; caml_formatVal v; Format.print_char ')'
  | TYv(ty) -> 
      Format.print_string "TYv("; caml_formatTy ty; Format.print_char ')'

let printToplevel s = Format.print_string s; Format.print_newline()

let caml_format prog =
  Format.print_string "{- Includes -}";
  Format.print_newline(); List.iter printToplevel prog.includes;
  Format.print_string "{- Static C Code -}";
  Format.print_newline(); List.iter printToplevel prog.static;
  Format.print_string "{- Processes -}";
  Format.print_newline(); caml_formatDec prog.dec; 
  Format.print_newline()

