(*
 * Fully-typed high-level Nomadic Pict syntax.
 *)

type status = RESPONSIVE | UNKNOWN

type proc =
  PRL of proc * proc
| IF of value * proc * proc
| INPUT of value * pat * proc
| OUTPUT of status * value * value
| INPA of value * pat * proc
| OUTA of status * value * value
| LDOUTA of status * value * value * value
| TRYHERE of value * value * proc * proc
| REP_INPUT of value * pat * proc
| TIMED_INPUT of value * value * pat * proc * proc
| REP_TIMED_INPUT of value * value * pat * proc * proc
| LI_SEND of value * value * value
| REMOTE_SEND of value * value * value * value
| LOCAL_SEND of value * value * value
| IFLOCAL of value * value * value * proc * proc
| LET of dec * proc
| SKIP
| TERMINATE
| TYPECASE of value * (pat * Types.t * proc) list * proc
| CASE of value * (Var.var * pat * proc) list

and pat =
  RECORDp of fieldPat list
| ASp of Var.var * ty * pat
| VARp of Var.var * ty
| WILDp

and fieldPat =
  ANONp of pat
| TYp of Var.var

and dec =
  DEF of (Var.var * pat * proc) list
| INLINE of Var.var * pat * proc
| NEW of Var.var * ty
| VAL of pat * value
| SEQ of dec * dec
| RUN of proc
| NEWLOC of  Var.var
| CREATE of (Var.var * proc) list
| CREATE_AGENT of (Var.var * proc) list
| AT of Inter.atom
| MIGRATE of value

and ty =
  BOXED
| UNBOXED
| VAR of Var.var

and value =
  CCODEv of Ccode.info * fieldVal list
| APPv of status * value * fieldVal list
| RECORDv of fieldVal list
| STRINGv of string
| SITEv of string * int
| SELECTv of value * int
| VARv of Inter.atom
| BOOLv of bool
| CHARv of char
| INTv of int
| LETv of dec * value
| IFv of value * value * value
| ABSv of pat * proc
| DYNAMICv of  value * Types.t
| TYPECASEv of value * (pat * Types.t * value) list * value
| VARIANTv of Var.var * value
| CASEv of value * (Var.var * pat * value) list

and fieldVal =
  ANONv of value
| TYv of ty

type toplevel =
  {includes : string list;
   static : string list;
   types : Inter.atom Types.c;
   dec : dec;
   context : Inter.atom Types.c}
    
val simplify : Inter.toplevel list -> toplevel -> Inter.toplevel





