(*
 * Syntax utilities.
 *)

type info = Error.info

type proc =
  PRL of info * proc * proc
| IF of info * value * proc * proc
| INPUT of info * value * pat * proc
| OUTPUT of info * value * value

| INPA of info * pat * proc
| OUTA of info * value * value
| LDOUTA of info * value * value * value
| TRYHERE of info * value * value * proc * proc

| REP_INPUT of info * value * pat * proc
| TIMED_INPUT of info * value * value * pat * proc * proc
| REP_TIMED_INPUT of info * value * value * pat * proc * proc
| LI_SEND of info * value * value * value
| REMOTE_SEND of info * value * value * value * value
| LOCAL_SEND of info * value * value * value
| IFLOCAL of info * value * value * value * proc * proc

| LET of info * dec * proc
| SKIP of info
| TERMINATE of info
| TYPECASE of info * value * (pat * proc) list * proc
| CASE of info * ty option * value * (string * pat * proc) list 
| ENCODE_PROC of info * string list

and pat =
  RECORDp of info * fieldPat list
| ASp of info * string * ty option * pat
| VARp of info * string * ty option
| WILDp of info * ty option
| RECp of info * ty option * pat
| ENCODE_PAT of info * string list

and fieldPat =
  ANONp of info * pat
| LABELEDp of info * string * pat
| TYp of info * string * constr

and dec =
  DEF of info * (info * string * pat * proc) list
| INLINE of info * string * pat * proc
| NEW of info * string * ty
| VAL of info * pat * value
| SEQ of info * dec * dec
| RUN of info * proc
| NEWLOC of info * string
| CREATE of info * (info * string * ty * proc) list
| CREATE_AGENT of info * (info * string * proc) list
| AT of info * string
| MIGRATE of info * value
| MACRO of info * string * value
| MACROLET of info * string * string * value
| DEFER_DEC of info * dec
| INLINE_DEC of info * dec
| INLINE_VAL of info * value
| DEF_TYPE of info * string * string list * ty
| DEF_DEF of info * string * string * pat * proc
| DEF_DEC of info * string * string list * dec
| DEF_PROC of info * string * string list * proc
| ENCODE_DEC of info * string list

and ty =
  TOP of info
| TVAR of info * string
| TABS of info * (string * Kind.kind) list * ty
| REC of info * string * Kind.kind * ty
| TAPP of info * ty * ty list
| RECORD of info * fieldTy list
| WITH of info * ty * fieldTy list
| WHERE of info * ty * fieldTy list
| CHAN of info * ty
| OCHAN of info * ty
| ICHAN of info * ty
| RCHAN of info * ty
| STRING of info
| BOOL of info
| CHAR of info
| INT of info
| SITE of info
| LOC of info
| AGENTT of info * ty
| AGENT of info
| DYN of info
| VART of info * (string * ty) list
| MAP of info * ty * ty
| ENCODE_TYPE of info * string list

and constr =
  LT of ty
| EQ of ty

and fieldTy =
  ANON of info * ty
| LABELED of info * string * ty
| TY of info * string * constr

and value =
  CCODEv of info * Ccode.info * fieldVal list
| APPv of info * ty option * value * fieldVal list
| RECORDv of info * fieldVal list
| WITHv of info * ty option * value * fieldVal list
| WHEREv of info * ty option * value * fieldVal list
| STRINGv of info * string
| SITEv of info * string * int
| SELECTv of info * value * string
| VARv of info * string
| BOOLv of info * bool
| CHARv of info * char
| INTv of info * int
| LETv of info * dec * value
| RECv of info * value * ty option
| IFv of info * ty option * value * value * value
| ABSv of info * pat * proc
| DYNAMICv of info * value * ty option
| TYPECASEv of info * ty option * value * (pat * value) list * value
| VARIANTv of info * string * value
| CASEv of info * ty option * value * (string * pat * value) list
| ENCODE_VALUE of info * string list

and fieldVal =
  ANONv of info * value
| LABELEDv of info * string * value
| TYv of info * ty

type toplevel =
  {imports : (Error.info * string) list;
   static : string list;
   includes : string list;
   dec : dec}

let procInfo = function
  PRL(i,_,_) -> i | IF(i,_,_,_) -> i | INPUT(i,_,_,_) -> i | OUTPUT(i,_,_) -> i 
| INPA(i,_,_) -> i | OUTA(i,_,_) -> i | LDOUTA(i,_,_,_) -> i | TRYHERE(i,_,_,_,_) -> i
| REP_INPUT(i,_,_,_) -> i | TIMED_INPUT(i,_,_,_,_,_) -> i | REP_TIMED_INPUT(i,_,_,_,_,_) -> i 
| LI_SEND(i,_,_,_) -> i | REMOTE_SEND(i,_,_,_,_) -> i 
| LOCAL_SEND(i,_,_,_) -> i | IFLOCAL(i,_,_,_,_,_) -> i 
| LET(i,_,_) -> i | SKIP(i) -> i
| TERMINATE(i) -> i | TYPECASE(i,_,_,_) -> i
| CASE(i,_,_,_) -> i | ENCODE_PROC(i,_) -> i

let patInfo = function
  RECORDp(i,_) -> i | ASp(i,_,_,_) -> i | VARp(i,_,_) -> i
| WILDp(i,_) -> i | RECp(i,_,_) -> i | ENCODE_PAT(i,_) -> i

let valInfo = function
  CCODEv(i,_,_) -> i | APPv(i,_,_,_) -> i | RECORDv(i,_) -> i
| STRINGv(i,_) -> i | SITEv(i,_,_) -> i | SELECTv(i,_,_) -> i | VARv(i,_) -> i
| BOOLv(i,_) -> i | CHARv(i,_) -> i | INTv(i,_) -> i | LETv(i,_,_) -> i
| RECv(i,_,_) -> i | IFv(i,_,_,_,_) -> i | WITHv(i,_,_,_) -> i
| WHEREv(i,_,_,_) -> i | ABSv(i,_,_) -> i
| DYNAMICv(i,_,_) -> i | TYPECASEv(i,_,_,_,_) -> i 
| VARIANTv(i,_,_) -> i | CASEv(i,_,_,_) -> i | ENCODE_VALUE(i,_) -> i

let fvInfo = function
  ANONv(i,_) -> i | LABELEDv(i,_,_) -> i | TYv(i,_) -> i

let decInfo = function
  DEF(i,_) -> i | INLINE(i,_,_,_) -> i | NEW(i,_,_) -> i
| VAL(i,_,_) -> i | SEQ(i,_,_) -> i | RUN(i,_) -> i
| NEWLOC(i,_) -> i | CREATE(i,_) -> i | CREATE_AGENT(i,_) -> i | AT(i,_) -> i | MIGRATE(i,_) -> i | MACRO(i,_,_) -> i | MACROLET(i,_,_,_) -> i
| DEFER_DEC(i,_) -> i | INLINE_DEC(i,_) -> i | INLINE_VAL(i,_) -> i 
| DEF_TYPE(i,_,_,_) -> i | DEF_DEF(i,_,_,_,_) -> i | DEF_DEC(i,_,_,_) -> i 
| DEF_PROC(i,_,_,_) -> i | ENCODE_DEC(i,_) -> i 

let typeInfo = function
  TVAR(i,_) -> i | TABS(i,_,_) -> i | TAPP(i,_,_) -> i | TOP(i) -> i
| REC(i,_,_,_) -> i | RECORD(i,_) -> i | WITH(i,_,_) -> i 
| RCHAN(i,_) -> i | OCHAN(i,_) -> i | STRING(i) -> i
| BOOL(i) -> i | CHAN(i,_) -> i | CHAR(i) -> i | INT(i) -> i
| ICHAN(i,_) -> i | WHERE(i,_,_) -> i
| SITE(i) -> i | LOC(i) -> i | AGENTT(i,_) -> i | AGENT(i) -> i | DYN(i) -> i
| VART(i,_) -> i | MAP(i,_,_) -> i | ENCODE_TYPE(i,_) -> i



